@@ -780,6 +780,18 @@ function onRemoveAccount(event) {
   else
     serverIndex++;
 
+  // Remove password information.
+  let serverUri = server.type + "://" + server.hostName;
+
+  let logins = Services.logins.findLogins({}, serverUri, null, serverUri);
+
+  for (let i = 0; i < logins.length; i++) {
+    if (logins[i].username == server.username) {
+      Services.logins.removeLogin(logins[i]);
+      break;
+    }
+  }
+
   try {
     let serverId = server.serverURI;
     MailServices.accounts.removeAccount(currentAccount);
@@ -1450,34 +1462,25 @@ var gAccountTree = {
   },
   onServerChanged: function at_onServerChanged(aServer) {},
 
-  _rdf: Components.classes["@mozilla.org/rdf/rdf-service;1"]
-                  .getService(Components.interfaces.nsIRDFService),
-  _rdfDataSource: null,
-  _rdfOpenAttribute: null,
+  _dataStore: Components.classes["@mozilla.org/xul/xulstore;1"]
+                        .getService(Components.interfaces.nsIXULStore),
 
   /**
-   * Retrieve from localstore.rdf whether the account should be expanded (open)
+   * Retrieve from XULStore.json whether the account should be expanded (open)
    * in the account tree.
    *
    * @param aAccountKey  key of the account to check
    */
   _getAccountOpenState: function at_getAccountOpenState(aAccountKey) {
-    // The code for this was ported from
-    // mozilla/browser/components/nsBrowserGlue.js.
-    if (!this._rdfDataSource) {
-      this._rdfDataSource = this._rdf.GetDataSource("rdf:local-store");
-      this._rdfOpenAttribute = this._rdf.GetResource("open");
-    }
-
-    // Retrieve the persisted value from localstore.rdf.
-    // It is stored under the URI of the current document and ID of the XUL element.
-    let resource = this._rdf.GetResource(document.documentURI + "#" + aAccountKey);
-    let target = this._rdfDataSource.GetTarget(resource, this._rdfOpenAttribute, true);
-    if (target instanceof Components.interfaces.nsIRDFLiteral)
-      return target.Value;
-
+    if (!this._dataStore.hasValue(document.documentURI, aAccountKey, "open")) {
     // If there was no value stored, use opened state.
-    return "true";
+      return "true";
+    } else {
+      // Retrieve the persisted value from XULStore.json.
+      // It is stored under the URI of the current document and ID of the XUL element.
+      return this._dataStore
+                 .getValue(document.documentURI, aAccountKey, "open");
+    }
   },
 
   _build: function at_build() {
@@ -1495,8 +1498,8 @@ var gAccountTree = {
 
     let mainTree = document.getElementById("account-tree-children");
     // Clear off all children...
-    while (mainTree.firstChild)
-      mainTree.removeChild(mainTree.firstChild);
+    while (mainTree.hasChildNodes())
+      mainTree.lastChild.remove();
 
     for (let account of accounts) {
       let accountName = null;
@@ -1591,9 +1594,9 @@ var gAccountTree = {
         }
         treeitem.setAttribute("container", "true");
         treeitem.id = accountKey;
-        // Load the 'open' state of the account from localstore.rdf.
+        // Load the 'open' state of the account from XULStore.json.
         treeitem.setAttribute("open", this._getAccountOpenState(accountKey));
-        // Let the localstore.rdf automatically save the 'open' state of the
+        // Let the XULStore.json automatically save the 'open' state of the
         // account when it is changed.
         treeitem.setAttribute("persist", "open");
       }
