@@ -101,7 +101,6 @@ class nsMsgDBView : public nsIMsgDBView, public nsIDBChangeListener,
 {
 public:
   nsMsgDBView();
-  virtual ~nsMsgDBView();
 
   NS_DECL_ISUPPORTS
   NS_DECL_NSIMSGDBVIEW
@@ -117,6 +116,8 @@ public:
                          class viewSortInfo *comparisonContext);
 
 protected:
+  virtual ~nsMsgDBView();
+
   static nsrefcnt gInstanceCount;
 
   static char16_t* kHighestPriorityString;
@@ -471,7 +472,7 @@ private:
                                          bool &changeReadState,
                                          nsIMsgFolder** targetFolder);
 
-  class nsMsgViewHdrEnumerator : public nsISimpleEnumerator 
+  class nsMsgViewHdrEnumerator MOZ_FINAL : public nsISimpleEnumerator
   {
   public:
     NS_DECL_ISUPPORTS
@@ -481,10 +482,12 @@ private:
 
     // nsMsgThreadEnumerator methods:
     nsMsgViewHdrEnumerator(nsMsgDBView *view);
-    ~nsMsgViewHdrEnumerator();
 
     nsRefPtr <nsMsgDBView> m_view;
     nsMsgViewIndex m_curHdrIndex;
+
+  private:
+    ~nsMsgViewHdrEnumerator();
   };
 };
 
