@@ -37,7 +37,7 @@
 // send it.
 const uint32_t kInitialMessageSendTime = 1000;
 
-NS_IMPL_ISUPPORTS7(nsMsgSendLater,
+NS_IMPL_ISUPPORTS(nsMsgSendLater,
                    nsIMsgSendLater,
                    nsIFolderListener,
                    nsIRequestObserver,
@@ -398,7 +398,7 @@ nsMsgSendLater::OnStartRequest(nsIRequest *request, nsISupports *ctxt)
 // This is the listener class for the send operation. We have to create this class 
 // to listen for message send completion and eventually notify the caller
 ////////////////////////////////////////////////////////////////////////////////////
-NS_IMPL_ISUPPORTS2(SendOperationListener, nsIMsgSendListener,
+NS_IMPL_ISUPPORTS(SendOperationListener, nsIMsgSendListener,
                    nsIMsgCopyServiceListener)
 
 SendOperationListener::SendOperationListener(nsMsgSendLater *aSendLater)
@@ -935,10 +935,6 @@ nsMsgSendLater::DeleteCurrentMessage()
 // representation.  Arguably, we could allow Lines to escape, but it's not
 // required by NNTP.)
 //
-#define UNHEX(C) \
-  ((C >= '0' && C <= '9') ? C - '0' : \
-  ((C >= 'A' && C <= 'F') ? C - 'A' + 10 : \
-        ((C >= 'a' && C <= 'f') ? C - 'a' + 10 : 0)))
 nsresult
 nsMsgSendLater::BuildHeaders()
 {
@@ -1095,14 +1091,10 @@ SEARCH_NEWLINE:
     }
     else if (do_flags_p)
     {
-      int i;
       char *s = value;
       PR_ASSERT(*s != ' ' && *s != '\t');
-      m_flags = 0;
-      for (i=0 ; i<4 ; i++) {
-      m_flags = (m_flags << 4) | UNHEX(*s);
-      s++;
-      }
+      NS_ASSERTION(MsgIsHex(s, 4), "Expected 4 hex digits for flags.");
+      m_flags = MsgUnhex(s, 4);
     }
 
     if (*buf == '\r' || *buf == '\n')
@@ -1511,7 +1503,7 @@ nsMsgSendLater::OnItemPropertyChanged(nsIMsgFolder *aItem, nsIAtom *aProperty,
 
 NS_IMETHODIMP
 nsMsgSendLater::OnItemIntPropertyChanged(nsIMsgFolder *aItem, nsIAtom *aProperty,
-                                         int32_t aOldValue, int32_t aNewValue)
+                                         int64_t aOldValue, int64_t aNewValue)
 {
   return NS_OK;
 }
