@@ -24,6 +24,7 @@
 #include "prmon.h"
 #include "nsIMsgImapMailFolder.h"
 #include "nsIMsgLocalMailFolder.h"
+#include "nsIMsgThread.h"
 #include "nsIImapMailFolderSink.h"
 #include "nsIImapServerSink.h"
 #include "nsIMsgFilterPlugin.h"
@@ -33,6 +34,7 @@
 #include "nsITimer.h"
 #include "nsCOMArray.h"
 #include "nsAutoSyncState.h"
+#include "nsIRequestObserver.h"
 
 class nsImapMoveCoalescer;
 class nsIMsgIdentity;
@@ -52,7 +54,6 @@ public:
     NS_DECL_THREADSAFE_ISUPPORTS
 
     nsImapMailCopyState();
-    virtual ~nsImapMailCopyState();
 
     nsCOMPtr<nsISupports> m_srcSupport; // source file spec or folder
     nsCOMPtr<nsIArray> m_messages; // array of source messages
@@ -84,6 +85,9 @@ public:
     // If the server supports UIDPLUS, this is the UID for the append,
     // if we're doing an append.
     nsMsgKey m_appendUID;
+
+private:
+    virtual ~nsImapMailCopyState();
 };
 
 NS_DEFINE_STATIC_IID_ACCESSOR(nsImapMailCopyState, NS_IMAPMAILCOPYSTATE_IID)
@@ -195,7 +199,6 @@ class nsImapMailFolder :  public nsMsgDBFolder,
  static const uint32_t PLAYBACK_TIMER_INTERVAL_IN_MS = 500; 
 public:
   nsImapMailFolder();
-  virtual ~nsImapMailFolder();
 
   NS_DECL_ISUPPORTS_INHERITED
 
@@ -229,7 +232,7 @@ public:
 
   NS_IMETHOD GetDeletable (bool *deletable) MOZ_OVERRIDE;
 
-  NS_IMETHOD GetSizeOnDisk(uint32_t * size) MOZ_OVERRIDE;
+  NS_IMETHOD GetSizeOnDisk(int64_t *size) MOZ_OVERRIDE;
 
   NS_IMETHOD GetCanCreateSubfolders(bool *aResult) MOZ_OVERRIDE;
   NS_IMETHOD GetCanSubscribe(bool *aResult) MOZ_OVERRIDE;
@@ -297,7 +300,6 @@ public:
   NS_DECL_NSIIMAPMAILFOLDERSINK
   NS_DECL_NSIIMAPMESSAGESINK
   NS_DECL_NSICOPYMESSAGELISTENER
-  NS_DECL_NSIREQUESTOBSERVER
 
   // nsIUrlListener methods
   NS_IMETHOD OnStartRunningUrl(nsIURI * aUrl) MOZ_OVERRIDE;
@@ -341,6 +343,7 @@ public:
   nsresult FindOpenRange(nsMsgKey &fakeBase, uint32_t srcCount);
 
 protected:
+  virtual ~nsImapMailFolder();
   // Helper methods
 
   virtual nsresult CreateChildFromURI(const nsCString &uri, nsIMsgFolder **folder) MOZ_OVERRIDE;
