@@ -5,33 +5,33 @@
 
 package org.mozilla.gecko.gfx;
 
+import java.nio.IntBuffer;
+import java.util.ArrayList;
+
+import org.mozilla.gecko.AndroidGamepadManager;
+import org.mozilla.gecko.AppConstants.Versions;
+import org.mozilla.gecko.EventDispatcher;
 import org.mozilla.gecko.GeckoAccessibility;
 import org.mozilla.gecko.GeckoAppShell;
 import org.mozilla.gecko.GeckoEvent;
 import org.mozilla.gecko.PrefsHelper;
-import org.mozilla.gecko.R;
 import org.mozilla.gecko.Tab;
 import org.mozilla.gecko.Tabs;
-import org.mozilla.gecko.TouchEventInterceptor;
 import org.mozilla.gecko.ZoomConstraints;
-import org.mozilla.gecko.mozglue.generatorannotations.WrapElementForJNI;
 import org.mozilla.gecko.mozglue.RobocopTarget;
-import org.mozilla.gecko.util.EventDispatcher;
+import org.mozilla.gecko.mozglue.generatorannotations.WrapElementForJNI;
 
 import android.content.Context;
 import android.graphics.Bitmap;
 import android.graphics.BitmapFactory;
 import android.graphics.Canvas;
 import android.graphics.Color;
-import android.graphics.PixelFormat;
 import android.graphics.Point;
 import android.graphics.PointF;
 import android.graphics.Rect;
 import android.graphics.SurfaceTexture;
-import android.os.Build;
 import android.os.Handler;
 import android.util.AttributeSet;
-import android.util.DisplayMetrics;
 import android.util.Log;
 import android.view.KeyEvent;
 import android.view.MotionEvent;
@@ -44,35 +44,34 @@ import android.view.inputmethod.EditorInfo;
 import android.view.inputmethod.InputConnection;
 import android.widget.FrameLayout;
 
-import java.nio.IntBuffer;
-import java.util.ArrayList;
-
 /**
  * A view rendered by the layer compositor.
  *
  * Note that LayerView is accessed by Robocop via reflection.
  */
 public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener {
-    private static String LOGTAG = "GeckoLayerView";
+    private static final String LOGTAG = "GeckoLayerView";
 
     private GeckoLayerClient mLayerClient;
     private PanZoomController mPanZoomController;
     private LayerMarginsAnimator mMarginsAnimator;
-    private GLController mGLController;
+    private final GLController mGLController;
     private InputConnectionHandler mInputConnectionHandler;
     private LayerRenderer mRenderer;
     /* Must be a PAINT_xxx constant */
     private int mPaintState;
     private int mBackgroundColor;
-    private boolean mFullScreen;
+    private FullScreenState mFullScreenState;
 
     private SurfaceView mSurfaceView;
     private TextureView mTextureView;
 
     private Listener mListener;
 
+    private PointF mInitialTouchPoint;
+    private boolean mGeckoIsReady;
+
     /* This should only be modified on the Java UI thread. */
-    private final ArrayList<TouchEventInterceptor> mTouchInterceptors;
     private final Overscroll mOverscroll;
 
     /* Flags used to determine when to show the painted surface. */
@@ -109,9 +108,9 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
         mGLController = GLController.getInstance(this);
         mPaintState = PAINT_START;
         mBackgroundColor = Color.WHITE;
+        mFullScreenState = FullScreenState.NONE;
 
-        mTouchInterceptors = new ArrayList<TouchEventInterceptor>();
-        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.ICE_CREAM_SANDWICH) {
+        if (Versions.feature14Plus) {
             mOverscroll = new OverscrollEdgeEffect(this);
         } else {
             mOverscroll = null;
@@ -139,18 +138,20 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
         setFocusableInTouchMode(true);
 
         GeckoAccessibility.setDelegate(this);
+        GeckoAccessibility.setAccessibilityStateChangeListener(getContext());
     }
 
-    private Point getEventRadius(MotionEvent event) {
-        if (Build.VERSION.SDK_INT >= 9) {
-            return new Point((int)event.getToolMajor()/2,
-                             (int)event.getToolMinor()/2);
-        }
+    /**
+     * MotionEventHelper dragAsync() robocop tests can instruct
+     * PanZoomController not to generate longpress events.
+     */
+    public void setIsLongpressEnabled(boolean isLongpressEnabled) {
+        ((JavaPanZoomController) mPanZoomController).setIsLongpressEnabled(isLongpressEnabled);
+    }
 
-        float size = event.getSize();
-        DisplayMetrics displaymetrics = getContext().getResources().getDisplayMetrics();
-        size = size * Math.min(displaymetrics.heightPixels, displaymetrics.widthPixels);
-        return new Point((int)size, (int)size);
+    private static Point getEventRadius(MotionEvent event) {
+        return new Point((int)event.getToolMajor() / 2,
+                         (int)event.getToolMinor() / 2);
     }
 
     public void geckoConnected() {
@@ -164,43 +165,36 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
         });
 
         mLayerClient.notifyGeckoReady();
-        addTouchInterceptor(new TouchEventInterceptor() {
-            private PointF mInitialTouchPoint = null;
+        mInitialTouchPoint = null;
+        mGeckoIsReady = true;
+    }
 
-            @Override
-            public boolean onInterceptTouchEvent(View view, MotionEvent event) {
-                return false;
-            }
+    private boolean sendEventToGecko(MotionEvent event) {
+        if (!mGeckoIsReady) {
+            return false;
+        }
 
-            @Override
-            public boolean onTouch(View view, MotionEvent event) {
-                if (event == null) {
-                    return true;
-                }
+        int action = event.getActionMasked();
+        PointF point = new PointF(event.getX(), event.getY());
+        if (action == MotionEvent.ACTION_DOWN) {
+            mInitialTouchPoint = point;
+        }
 
-                int action = event.getActionMasked();
-                PointF point = new PointF(event.getX(), event.getY());
-                if (action == MotionEvent.ACTION_DOWN) {
-                    mInitialTouchPoint = point;
-                }
+        if (mInitialTouchPoint != null && action == MotionEvent.ACTION_MOVE) {
+            Point p = getEventRadius(event);
 
-                if (mInitialTouchPoint != null && action == MotionEvent.ACTION_MOVE) {
-                    Point p = getEventRadius(event);
-
-                    if (PointUtils.subtract(point, mInitialTouchPoint).length() <
-                        Math.max(PanZoomController.CLICK_THRESHOLD, Math.min(Math.min(p.x, p.y), PanZoomController.PAN_THRESHOLD))) {
-                        // Don't send the touchmove event if if the users finger hasn't moved far.
-                        // Necessary for Google Maps to work correctly. See bug 771099.
-                        return true;
-                    } else {
-                        mInitialTouchPoint = null;
-                    }
-                }
-
-                GeckoAppShell.sendEventToGecko(GeckoEvent.createMotionEvent(event, false));
+            if (PointUtils.subtract(point, mInitialTouchPoint).length() <
+                Math.max(PanZoomController.CLICK_THRESHOLD, Math.min(Math.min(p.x, p.y), PanZoomController.PAN_THRESHOLD))) {
+                // Don't send the touchmove event if if the users finger hasn't moved far.
+                // Necessary for Google Maps to work correctly. See bug 771099.
                 return true;
+            } else {
+                mInitialTouchPoint = null;
             }
-        });
+        }
+
+        GeckoAppShell.sendEventToGecko(GeckoEvent.createMotionEvent(event, false));
+        return true;
     }
 
     public void showSurface() {
@@ -223,37 +217,6 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
         Tabs.unregisterOnTabsChangedListener(this);
     }
 
-    public void addTouchInterceptor(final TouchEventInterceptor aTouchInterceptor) {
-        post(new Runnable() {
-            @Override
-            public void run() {
-                mTouchInterceptors.add(aTouchInterceptor);
-            }
-        });
-    }
-
-    public void removeTouchInterceptor(final TouchEventInterceptor aTouchInterceptor) {
-        post(new Runnable() {
-            @Override
-            public void run() {
-                mTouchInterceptors.remove(aTouchInterceptor);
-            }
-        });
-    }
-
-    private boolean runTouchInterceptors(MotionEvent event, boolean aOnTouch) {
-        boolean result = false;
-        for (TouchEventInterceptor i : mTouchInterceptors) {
-            if (aOnTouch) {
-                result |= i.onTouch(this, event);
-            } else {
-                result |= i.onInterceptTouchEvent(this, event);
-            }
-        }
-
-        return result;
-    }
-
     @Override
     public void dispatchDraw(final Canvas canvas) {
         super.dispatchDraw(canvas);
@@ -270,28 +233,25 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
             requestFocus();
         }
 
-        if (runTouchInterceptors(event, false)) {
+        if (mMarginsAnimator != null && mMarginsAnimator.onInterceptTouchEvent(event)) {
             return true;
         }
         if (mPanZoomController != null && mPanZoomController.onTouchEvent(event)) {
             return true;
         }
-        if (runTouchInterceptors(event, true)) {
-            return true;
-        }
-        return false;
+        return sendEventToGecko(event);
     }
 
     @Override
     public boolean onHoverEvent(MotionEvent event) {
-        if (runTouchInterceptors(event, true)) {
-            return true;
-        }
-        return false;
+        return sendEventToGecko(event);
     }
 
     @Override
     public boolean onGenericMotionEvent(MotionEvent event) {
+        if (AndroidGamepadManager.handleMotionEvent(event)) {
+            return true;
+        }
         if (mPanZoomController != null && mPanZoomController.onMotionEvent(event)) {
             return true;
         }
@@ -325,12 +285,13 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
 
             SurfaceHolder holder = mSurfaceView.getHolder();
             holder.addCallback(new SurfaceListener());
-            holder.setFormat(PixelFormat.RGB_565);
         }
     }
 
-    @RobocopTarget
-    public GeckoLayerClient getLayerClient() { return mLayerClient; }
+    // Don't expose GeckoLayerClient to things outside this package; only expose it as an Object
+    GeckoLayerClient getLayerClient() { return mLayerClient; }
+    public Object getLayerClientObject() { return mLayerClient; }
+
     public PanZoomController getPanZoomController() { return mPanZoomController; }
     public LayerMarginsAnimator getLayerMarginsAnimator() { return mMarginsAnimator; }
 
@@ -368,7 +329,6 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
 
     public void setInputConnectionHandler(InputConnectionHandler inputConnectionHandler) {
         mInputConnectionHandler = inputConnectionHandler;
-        mLayerClient.forceRedraw(null);
     }
 
     @Override
@@ -523,13 +483,8 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
      * TextureView instead of a SurfaceView, the first phase is skipped.
      */
     private void onSizeChanged(int width, int height) {
-        if (!mGLController.isCompositorCreated()) {
-            return;
-        }
-
-        surfaceChanged(width, height);
-
-        if (mSurfaceView == null) {
+        if (!mGLController.isServerSurfaceValid() || mSurfaceView == null) {
+            surfaceChanged(width, height);
             return;
         }
 
@@ -645,11 +600,24 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
         }
     }
 
+    @RobocopTarget
+    public void addDrawListener(DrawListener listener) {
+        mLayerClient.addDrawListener(listener);
+    }
+
+    @RobocopTarget
+    public void removeDrawListener(DrawListener listener) {
+        mLayerClient.removeDrawListener(listener);
+    }
+
+    @RobocopTarget
+    public static interface DrawListener {
+        public void drawFinished();
+    }
+
     @Override
     public void setOverScrollMode(int overscrollMode) {
-        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.GINGERBREAD) {
-            super.setOverScrollMode(overscrollMode);
-        }
+        super.setOverScrollMode(overscrollMode);
         if (mPanZoomController != null) {
             mPanZoomController.setOverScrollMode(overscrollMode);
         }
@@ -661,10 +629,7 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
             return mPanZoomController.getOverScrollMode();
         }
 
-        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.GINGERBREAD) {
-            return super.getOverScrollMode();
-        }
-        return View.OVER_SCROLL_ALWAYS;
+        return super.getOverScrollMode();
     }
 
     @Override
@@ -673,12 +638,16 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
         GeckoAccessibility.onLayerViewFocusChanged(this, gainFocus);
     }
 
-    public void setFullScreen(boolean fullScreen) {
-        mFullScreen = fullScreen;
+    public void setFullScreenState(FullScreenState state) {
+        mFullScreenState = state;
     }
 
     public boolean isFullScreen() {
-        return mFullScreen;
+        return mFullScreenState != FullScreenState.NONE;
+    }
+
+    public FullScreenState getFullScreenState() {
+        return mFullScreenState;
     }
 
     @Override
@@ -688,4 +657,15 @@ public class LayerView extends FrameLayout implements Tabs.OnTabsChangedListener
             setIsRTL(tab.getIsRTL());
         }
     }
+
+    // Public hooks for listening to metrics changing
+
+    public interface OnMetricsChangedListener {
+        public void onMetricsChanged(ImmutableViewportMetrics viewport);
+        public void onPanZoomStopped();
+    }
+
+    public void setOnMetricsChangedListener(OnMetricsChangedListener listener) {
+        mLayerClient.setOnMetricsChangedListener(listener);
+    }
 }
