@@ -5,13 +5,13 @@
 
 package org.mozilla.gecko.home;
 
-import org.mozilla.gecko.db.BrowserContract.Combined;
 import org.mozilla.gecko.util.StringUtils;
 
+import android.database.Cursor;
 import android.text.TextUtils;
 import android.view.View;
 import android.widget.AdapterView.AdapterContextMenuInfo;
-
+import android.widget.ExpandableListAdapter;
 
 /**
  * A ContextMenuInfo for HomeListView
@@ -20,11 +20,16 @@ public class HomeContextMenuInfo extends AdapterContextMenuInfo {
 
     public String url;
     public String title;
-    public boolean isFolder = false;
-    public boolean inReadingList = false;
-    public int display = Combined.DISPLAY_NORMAL;
+    public boolean isFolder;
     public int historyId = -1;
     public int bookmarkId = -1;
+    public int readingListItemId = -1;
+    public RemoveItemType itemType = null;
+
+    // Item type to be handled with "Remove" selection.
+    public static enum RemoveItemType {
+        BOOKMARKS, HISTORY, READING_LIST
+    }
 
     public HomeContextMenuInfo(View targetView, int position, long id) {
         super(targetView, position, id);
@@ -39,7 +44,11 @@ public class HomeContextMenuInfo extends AdapterContextMenuInfo {
     }
 
     public boolean isInReadingList() {
-        return inReadingList;
+        return readingListItemId > -1;
+    }
+
+    public boolean canRemove() {
+        return hasBookmarkId() || hasHistoryId() || isInReadingList();
     }
 
     public String getDisplayTitle() {
@@ -48,4 +57,18 @@ public class HomeContextMenuInfo extends AdapterContextMenuInfo {
         }
         return StringUtils.stripCommonSubdomains(StringUtils.stripScheme(url, StringUtils.UrlFlags.STRIP_HTTPS));
     }
-}
\ No newline at end of file
+
+    /**
+     * Interface for creating ContextMenuInfo instances from cursors.
+     */
+    public interface Factory {
+        public HomeContextMenuInfo makeInfoForCursor(View view, int position, long id, Cursor cursor);
+    }
+
+    /**
+     * Interface for creating ContextMenuInfo instances from ExpandableListAdapters.
+     */
+    public interface ExpandableFactory {
+        public HomeContextMenuInfo makeInfoForAdapter(View view, int position, long id, ExpandableListAdapter adapter);
+    }
+}
